from difflib import SequenceMatcher
from inspect import getmembers, ismethod
from typing import Callable, Dict, List, Tuple

from theeng.core.problem import ProblemConstructor


class Step:
    def __init__(
        self,
        problem: ProblemConstructor,
        evaluator: Callable[[Dict[str, float]], Dict[str, float]],
    ):
        self.problem = problem
        self.evaluator = evaluator

        pNames = problem.getPnames()
        objectiveExpressions = problem.getObjectivesExpressions()
        resultsExpressions = problem.getResultsExpressions()

        Step._checkExpressions(pNames, "Parameters Names")
        Step._checkExpressions(objectiveExpressions, "Objectives Names")
        Step._checkExpressions(resultsExpressions, "Results Names")

        self.pNames = pNames
        self.objectives = problem.getObjectives()
        self.constraints = problem.getConstraints()
        self.objectiveExpressions = objectiveExpressions
        self.constraintExpressions = problem.getConstraintsExpressions()
        self.resultsExpressions = resultsExpressions
        self.lowerBounds, self.upperBounds = problem.getBounds()

        self.nVar = len(self.pNames)
        self.nObj = len(self.objectiveExpressions)

    def _getMethod(self, classObject: Callable, methodName: str, **kwargs):
        obj = classObject(**kwargs)
        availableMethods = [m[0] for m in getmembers(obj, predicate=ismethod)]
        if methodName not in availableMethods:
            similarMethods, similarity_ratio = Step._findSimilar(
                methodName, availableMethods
            )
            similarMethod = similarMethods[0]
            print(
                f"Method {methodName} not available or misspelled. Using {similarMethod} instead.\nMatching percentage: {round(similarity_ratio[0]*100, 2)} %"
            )
            methodName = similarMethod  # overwrite method with similar method
        samplerMethod = getattr(obj, methodName)
        return samplerMethod

    @staticmethod
    def _findSimilar(
        source_name: str, nameslist: List[str]
    ) -> Tuple[List[str], List[float]]:
        """
        Find similar names to a source name in a list of names.

        Args:
            source_name (str): the name to compare.
            nameslist (List[str]): the list of names to compare with the source name.

        Returns:
            Tuple[List[str], List[float]]: A tuple of sorted lists of similar names and their similarities.
        """
        similar_names_similarity = []
        for name in nameslist:
            similarity = SequenceMatcher(None, source_name, name).ratio()
            if similarity > 0.8:
                similar_names_similarity.append((name, similarity))

        similar_names_similarity.sort(key=lambda x: x[1], reverse=True)
        similar_names = [name for name, _ in similar_names_similarity]
        similarities = [similarity for _, similarity in similar_names_similarity]

        return similar_names, similarities

    @staticmethod
    def _checkExpressions(expressions: List[str], listName: str):
        if not expressions:
            raise ValueError(f"List of {listName} is empty.")



from os.path import isfile
from pickle import load
from typing import List

import streamlit as st


class SurrogateView:
    def __init__(self, path: str) -> None:
        st.set_page_config(layout="wide")
        st.title("Surrogate Inspector")
        self.surrogatePath = None
        self.pNames = None
        self.resultNames = None
        self.names = None
        self.designEvaluation = []
        self.path = path

        with st.expander("Settings", expanded=True):
            left, center, right = st.columns((1, 1, 1))
            with left:
                evaluator = self._setSurrogatePath(self.path)
            with center:
                self._setParameters()
            with right:
                self._setTargets()

            self.names = self.pNames + self.resultNames  # type: ignore

        with st.container():
            left_1, right_1 = st.columns((1, 1))
            with left_1:
                numericInputs = self._getInputs(self.pNames)  # type: ignore
                predictButton = st.button("Predict")
            with right_1:
                if predictButton:
                    if evaluator:
                        resultList = evaluator(numericInputs)  # type: ignore
                        self._getOutputs(self.resultNames, resultList)  # type: ignore

    def _setSurrogatePath(self, path: str):
        self.surrogatePath = st.text_input("Path to surrogate", path)
        if isfile(self.surrogatePath):
            with open(self.surrogatePath, "rb") as surrogateFile:
                surrogate = load(surrogateFile)  # type: ignore

            def evaluator(parameterValues):
                return surrogate.predict([parameterValues])

            return evaluator
        else:
            st.error("Surrogate file does not exist. Check path and filename.")
            return None

    def _setParameters(self):
        parametersString = st.text_input(
            "Parameters name"
        )  # todo: use parameters from main script if available as default
        self.pNames = parametersString.replace(" ", "").split(",")

    def _setTargets(self):
        resultNamesString = st.text_input(
            "Results name"
        )  # todo: use results from main script if available as default
        self.resultNames = resultNamesString.replace(" ", "").split(",")

    def _getInputs(self, inputsList: List[str]) -> List[float]:
        inputsList = list(filter(None, inputsList))
        numericInputs = []
        for count, input in enumerate(inputsList):
            numericInputs.append(st.number_input(input, key=count))
        return numericInputs

    def _getOutputs(self, outputsList: List[str], resultList: List[List[float]]):
        outputsList = list(filter(None, outputsList))
        textOutputs = []
        for output in zip(outputsList, resultList[0]):
            textOutputs.append(st.metric(output[0], output[1]))


if __name__ == "__main__":
    view = SurrogateView(
        path="C:\\Repositories\\TheEng\\examples\\beam_freecad_multiobj\\surrogate.pkl"
    )



from os.path import isfile
from typing import Callable, Dict

from theeng.algorithms.simulators import Simulators
from theeng.core.abstract import Step
from theeng.core.problem import ProblemConstructor


class Simulator(Step):
    def __init__(self, problem: ProblemConstructor) -> None:
        self.resultsExpressions = problem.getResultsExpressions()
        self.iterableOutput = problem.getIterableOutput()
        self.simulator = None

    def generate(
        self, simulatorName: str, fcdPath: str
    ) -> Callable[[Dict[str, float]], Dict[str, float]]:
        if not isfile(fcdPath):
            raise FileNotFoundError(
                f"FreeCAD file at {fcdPath} was not found. Check path or filename."
            )
        simulator = self._getMethod(
            Simulators,
            simulatorName,
            resultsExpressions=self.resultsExpressions,
            iterableOutput=self.iterableOutput,
            fcdPath=fcdPath,
        )
        self.simulator = simulator
        return simulator

    def simulate(self, parameters: Dict[str, float]) -> Dict[str, float]:
        if not self.simulator:
            raise ValueError("No simulator has been generated. Use do() method first.")
        return self.simulator(parameters)



from collections import defaultdict
from typing import Callable, Dict, List, Tuple

from numpy import concatenate
from pandas import DataFrame
from scipy.stats import qmc

from theeng.algorithms.samplers import Samplers
from theeng.core.abstract import Step
from theeng.core.problem import ProblemConstructor


class Sampler(Step):
    def __init__(
        self,
        problem: ProblemConstructor,
        evaluator: Callable[[Dict[str, float]], Dict[str, float]],
    ) -> None:
        super().__init__(problem, evaluator)

    def sample(
        self, samplerName: str = "latinHypercube", nSamples: int = 50
    ) -> Tuple[List[List[float]], List[List[float]], DataFrame]:
        problem = SamplingProblem(self.problem, self.evaluator)
        samplerMethod = self._getMethod(Samplers, samplerName, nVar=self.nVar)()

        samp = samplerMethod.random(n=nSamples)
        x = qmc.scale(samp, self.lowerBounds, self.upperBounds).tolist()

        out = defaultdict(list)
        res = problem._evaluate(x, out)  # type: ignore

        f = res["F"]
        r = res["R"]

        data = concatenate([x, r], axis=1)
        data = DataFrame(
            data,
            columns=self.pNames
            + self.resultsExpressions
            + self.objectiveExpressions
            + self.constraintExpressions,
        )

        data = data.T.drop_duplicates().T

        return x, f, data


class SamplingProblem:
    def __init__(
        self,
        problem: ProblemConstructor,
        evaluator: Callable[[Dict[str, float]], Dict[str, float]],
    ):
        """Initialize the sampling problem.

        Args:
            problem (ProblemConstructor): The problem to be evaluated.
            evaluator (Evaluator): The evaluator to be used.
        """
        self._evaluator = evaluator

        self._nVar = problem.getNvar()
        self._pNames = problem.getPnames()

        self._objectives = problem.getObjectives()
        self._constraints = problem.getConstraints()
        self._lowerBounds, self._upperBounds = problem.getBounds()

    def _evaluate(
        self, x: List[List[float]], out: Dict[str, List[List[float]]], *args, **kwargs
    ) -> Dict[str, List[List[float]]]:
        """Evaluate the sampling problem on the given samples.

        Args:
            x (List[List[float]]): Parameters samples.
            out (Dict[str, List[List[float]]]): Retruned dictionary (inspired by Pymoo).

        Returns:
            Dict[str, List[List[float]]]: The evaluated samples, objectives and constraints.
        """
        f = []
        g = []
        r = []

        for sample in x:
            parameters = {name: value for name, value in zip(self._pNames, sample)}
            results = self._evaluator(parameters)

            objs = [obj(results) for obj in self._objectives]
            consts = [constr(results) for constr in self._constraints]
            res = list(results.values()) + objs + consts

            f.append(objs)
            g.append(consts)
            r.append(res)

        out["F"] = f
        out["G"] = g
        out["R"] = r

        return out



from collections import defaultdict
from json import load
from os.path import exists
from sys import path
from typing import Dict, Iterable, List, Union

from numpy import average, max, min

f = open("configs\\settings.json")
data = load(f)
freeCADBinPath = data["FREECAD_PATH"]
if not exists(freeCADBinPath):
    raise FileNotFoundError(
        "provided FreeCAD binary directory does not exists. Check FreeCAD installation path in settings.json file."
    )

path.append(data["FREECAD_PATH"])

import FreeCAD
from femtools import ccxtools


class Simulators:
    def __init__(
        self,
        resultsExpressions: List[str],
        iterableOutput: List[Union[str, None]],
        fcdPath: str,
    ) -> None:
        """Initialize an FEM evaluator.

        Args:
            problem (ProblemConstructor): problem to be evaluated.
            resultsRequest (List[str]): list of results aliases contained in the spreadsheet.
            fcdPath (str): path to the FreeCAD file containing the model.
        """
        self.resultsExpressions = resultsExpressions
        self.iterableOutput = iterableOutput
        self._doc = FreeCAD.open(fcdPath)
        self._sheet = self._doc.getObject("Spreadsheet")

    def femSimulator(self, parameters: Dict[str, float]) -> Dict[str, float]:
        """Evaluate the design parameters and return the results by updating the spreadsheet and running the FEM analysis in FreeCAD.

        Args:
            parameters (Dict[str, float]): dictionary of design parameters containing aliases and values contained in the spreadsheet.

        Returns:
            Dict[str, float]: dictionary containing results aliases and values.
        """

        for key, value in parameters.items():
            self._sheet.set(key, str(value))

        # self._sheet.recompute()
        self._doc.recompute()

        fea = ccxtools.FemToolsCcx()
        fea.update_objects()
        fea.setup_working_dir()
        fea.setup_ccx()
        message = fea.check_prerequisites()
        if not message:
            fea.purge_results()
            fea.write_inp_file()
            fea.ccx_run()
            fea.load_results()
        else:
            print(
                "Houston, we have a problem! {}\n".format(message)
            )  # in Python console

        self._sheet.recompute()
        results = defaultdict(float)
        for result, iterableAction in zip(self.resultsExpressions, self.iterableOutput):
            if result in parameters:
                ccx_result = parameters[result]
            else:
                ccx_result = self._sheet.get(result)
            if iterableAction is None:
                if not isinstance(ccx_result, float):
                    raise Exception("Result is not float.")
                results[result] = ccx_result
            elif iterableAction == "Max":
                if not isinstance(ccx_result, Iterable):
                    raise Exception("Result is not Iterable.")
                results[result] = max(ccx_result)  # type: ignore
            elif iterableAction == "Min":
                if not isinstance(ccx_result, Iterable):
                    raise Exception("Result is not Iterable.")
                results[result] = min(ccx_result)  # type: ignore
            elif iterableAction == "Avg":
                if not isinstance(ccx_result, Iterable):
                    raise Exception("Result is not Iterable.")
                results[result] = average(ccx_result)  # type: ignore
            else:
                raise Exception("Invalid iterable action.")

        return results

    def cfdSimulator(self, parameters: Dict[str, float]) -> Dict[str, float]:
        raise NotImplementedError("CFD interface is not implemented yet.")



from typing import Tuple

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import PolynomialFeatures, SplineTransformer, StandardScaler
from sklearn.svm import SVR


class Surrogates:
    def __init__(self) -> None:
        pass

    def polynomial(
        self,
        degree_fit: int = 2,
        interaction_only: bool = False,
        fit_intercept: bool = True,
        **kwargs
    ) -> Pipeline:
        """_summary_

        Args:
            degree_fit (int, optional): The degree of the fitting polynomial. Defaults to 2.
            interaction_only (bool, optional): If to consider design variable interactions. Defaults to False.
            fit_intercept (bool, optional): Whether to fit the intercept. Defaults to True.

        Returns:
            Pipeline: A Scikit-Learn pipeline.
        """
        pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "poly",
                    PolynomialFeatures(
                        degree=degree_fit,
                        interaction_only=interaction_only,
                    ),
                ),
                (
                    "linear",
                    LinearRegression(fit_intercept=fit_intercept),
                ),
            ]
        )

        return pipeline

    def gaussianProcess(self, kernel=None, **kwargs) -> Pipeline:
        """_summary_

        Args:
            kernel (_type_, optional): Kernel for the Gauss process. Defaults to None.

        Returns:
            Pipeline: A Scikit-Learn pipeline.
        """
        pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "gauss",
                    GaussianProcessRegressor(kernel=kernel, random_state=0),
                ),
            ]
        )

        return pipeline

    def neuralNetwork(
        self,
        n_nodes: Tuple[int, int] = (16, 8),
        activation: str = "relu",
        solver: str = "adam",
        n_epochs: int = 1000,
        **kwargs
    ) -> Pipeline:
        """_summary_

        Args:
            n_nodes (Tuple[int, int], optional): NUmber of nodes in each layer. Defaults to (16, 8).
            activation (str, optional): Activation function for the NN. Defaults to "relu".
            solver (str, optional): The backpropagation optimizer. Defaults to "adam".
            n_epochs (int, optional): The number of epochs. Defaults to 1000.

        Returns:
            Pipeline: A Scikit-Learn pipeline.
        """
        pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "mlp",
                    MLPRegressor(
                        hidden_layer_sizes=n_nodes,
                        activation=activation,
                        solver=solver,
                        max_iter=n_epochs,
                        early_stopping=True,
                    ),
                ),
            ]
        )

        return pipeline

    def supportVector(
        self, kernel: str = "rbf", degree_fit: int = 2, **kwargs
    ) -> Pipeline:
        """_summary_

        Args:
            kernel (str, optional): The Kernel of the SVM. Defaults to "rbf".
            degree_fit (int, optional): The degree of the fitting polynomial. Defaults to 2.

        Returns:
            Pipeline: A Scikit-Learn pipeline.
        """
        pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "svr",
                    SVR(
                        kernel=kernel,
                        degree=degree_fit,
                    ),
                ),
            ]
        )

        return pipeline

    def spline(
        self,
        n_knots: int = 2,
        degree_fit: int = 2,
        fit_intercept: bool = True,
        **kwargs
    ) -> Pipeline:
        """_summary_

        Args:
            n_knots (int, optional): Number of spline knots. Defaults to 2.
            degree_fit (int, optional): The degree of the fitting polynomial. Defaults to 2.
            fit_intercept (bool, optional): Whether to fit the intercept. Defaults to True.

        Returns:
            Pipeline: A Scikit-Learn pipeline.
        """
        pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "spline",
                    SplineTransformer(
                        n_knots=n_knots,
                        degree=degree_fit,
                    ),
                ),
                (
                    "linear",
                    LinearRegression(fit_intercept=fit_intercept),
                ),
            ]
        )

        return pipeline



import json
from os.path import isfile, join

from pandas import concat

from theeng.core.optimizer import Optimizer
from theeng.core.problem import ProblemConstructor
from theeng.core.ranker import Ranker
from theeng.core.sampler import Sampler
from theeng.core.simulator import Simulator
from theeng.core.surrogate import Surrogate
from theeng.core.visualization import Visualization


class TheEng:
    def __init__(
        self
    ) -> None:

        self.makeSurrogate = None

        self.settingsReady = False

        self.workingDirectory = ""
        self.simulationDirectory = ""
        self.simulatorName = ""
        self.nCPUs = None
        self.results = None
        self.bounds = None
        self.samplerName = ""
        self.nSamples = None
        self.surrogateName = ""
        self.degree_fit = None
        self.fit_intercept = None
        self.fit_interactions = None
        self.optimizerName = ""
        self.popSize = None
        self.termination = None
        self.rankingName = ""
        self.objectives = None
        self.objectiveWeights = None
        self.constraints = None
        self.constraintsRelaxations = None

    def run(self):
        if not self.settingsReady:
            raise Exception(
                "Settings are not ready. Please run getSettingsFromDict() or getSettingsFromJson() method first."
            )
        problem = ProblemConstructor()
        problem.setResults(self.results)  # type: ignore
        problem.setObjectives(self.objectives)  # type: ignore
        problem.setContraints(self.constraints)  # type: ignore
        problem.setBounds(self.bounds)  # type: ignore

        simul = Simulator(problem)
        simulator = simul.generate(
            simulatorName=self.simulatorName,
            fcdPath=self.simulationDirectory,
        )

        evaluator = simulator

        if self.makeSurrogate:
            sampler = Sampler(problem, simulator)
            _, _, dataSamp = sampler.sample(nSamples=self.nSamples)  # type: ignore

            surrog = Surrogate(problem, dataSamp)  # type: ignore
            surrogate, _ = surrog.generate(
                surrogateName=self.surrogateName,
                save=True,
                degree_fit=self.degree_fit,
                surrogatePath=join(self.workingDirectory, "surrogate.pkl"),
            )
            evaluator = surrogate

        optimizer = Optimizer(problem, evaluator)
        xOpt, _, dataOpt = optimizer.optimize(
            optimizerName=self.optimizerName, termination=self.termination, popSize=self.popSize  # type: ignore
        )

        if self.makeSurrogate:
            _, _, dataOpt = optimizer.convertToSimulator(xOpt, simulator)
            data = concat([dataSamp, dataOpt])
        else:
            data = dataOpt

        ranker = Ranker(problem, data)
        dataRanked = ranker.rank(rankingName=self.rankingName)

        dataRanked.to_csv(join(self.workingDirectory, "db.csv"))

        visualizer = Visualization(dataRanked)

        visualizer.plot(
            visualizationName="parallelCoordinate",
            savePath=join(self.workingDirectory, "parallel_coord.html"),
        )
        visualizer.plot(
            visualizationName="heatMap",
            savePath=join(self.workingDirectory, "heatmap.html"),
        )

    def getSettingsFromJson(self, jsonPath):
        self._checkFileExists(jsonPath)
        with open(jsonPath, "r") as f:
            self._settings = json.load(f)
        self.getSettingsFromDict(self._settings)

    def getSettingsFromDict(self, settings):
        self._settings = settings

        self._getGeneralSettings()
        self._getProblemSettings()
        if self.makeSurrogate:
            self._getSamplingSettings()
            self._getSurrogateSettings()
        self._getOptimizationSettings()

        self.settingsReady = True

    def _getGeneralSettings(self):
        generalSettings = self._settings["General Settings"]
        self.workingDirectory = generalSettings["Working Directory"]
        self.simulationDirectory = generalSettings["Simulation Directory"]
        self.simulatorName = generalSettings["Simulator Name"]
        self.makeSurrogate = generalSettings["Use Surrogate"]
        self.nCPUs = generalSettings["nCPUs"]

    def _getProblemSettings(self):
        problemSettings = self._settings["Problem"]
        self.bounds = problemSettings["Parameters"]
        self.results = {}
        for resultName, resultOperation in problemSettings["Results"].items():
            if resultOperation == "None":
                resultOperation = None
            self.results[resultName] = resultOperation

    def _getSamplingSettings(self):
        samplingSettings = self._settings["Sampling"]
        self.samplerName = samplingSettings["Method"]
        self.nSamples = samplingSettings["Number of Samples"]

    def _getSurrogateSettings(self):
        surrogateSettings = self._settings["Surrogate"]
        self.surrogateName = surrogateSettings["Method"]
        self.degree_fit = surrogateSettings["Degree of Fit"]
        self.fit_intercept = surrogateSettings["Fit Intercept"]
        self.fit_interactions = surrogateSettings["Fit Interactions"]

    def _getOptimizationSettings(self):
        optimizationSettings = self._settings["Optimization"]
        self.optimizerName = optimizationSettings["Method"]
        self.popSize = optimizationSettings["Population Size"]
        n_eval = optimizationSettings["Number of Evaluations"]
        self.termination = ("n_eval", n_eval)
        self.rankingName = optimizationSettings["Ranking Method"]
        self.objectives = optimizationSettings["Objectives Expressions"]
        self.constraints = optimizationSettings["Constraints Expressions"]

    # check if file exists
    def _checkFileExists(self, filePath):
        if not isfile(filePath):
            raise FileNotFoundError(
                "Input json file does not exist. Check path and try again."
            )



import operator
from typing import Callable, Dict, Iterable, List, Tuple, Union


class ProblemConstructor:
    """The ProblemConstructor class is used to construct the problem to be solved by the optimizer."""

    def __init__(self) -> None:
        """Initialize the ProblemConstructor class."""

        self._objectives = []
        self.objectivesExpressions = []
        self._loweBounds = []
        self._upperBounds = []
        self.boundsExpressions = []
        self._constraints = []
        self.constraintsExpressions = []
        self.resultsExpressions = []
        self.iterableOutput = []
        self.objectiveWeights = []
        self.constraintsRelaxation = []

        self.nobj = 0
        self.nconst = 0
        self.nvar = 0
        self.pnames = []

    def setObjectives(self, expressions: Dict[str, float]) -> None:
        """Set the objectives of the problem.

        Args:
            expressions (Dict[str, float]): Dict of expressions encoded as strings with their weights.
        """

        ProblemConstructor._checkExpressions(expressions)

        for expression in expressions.keys():
            operands, operations = ProblemConstructor._expressionParser(expression)
            self._objectives.append(
                lambda results, operands=operands, operations=operations: ProblemConstructor._expressionEvaluator(
                    results, operands, operations
                )
            )
            self.nobj += 1
        self.objectivesExpressions = list(expressions.keys())
        self.objectiveWeights = list(expressions.values())

    def setContraints(self, expressions: Dict[str, float]) -> None:
        """Set the constraints of the problem.

        Args:
            expressions (Dict[str, float]): Dictionary of expressions encoded as strings with the allowable relaxation. They will be considered <= 0.
        """

        ProblemConstructor._checkExpressions(expressions)

        for expression in expressions.keys():
            operands, operations = ProblemConstructor._expressionParser(expression)
            self._constraints.append(
                lambda results, operands=operands, operations=operations: ProblemConstructor._expressionEvaluator(
                    results, operands, operations
                )
            )
            self.nconst += 1
        self.constraintsExpressions = list(expressions.keys())
        self.constraintsRelaxation = list(expressions.values())

    def setBounds(self, bounds: Dict[str, Tuple[float, float]]) -> None:
        """Set the bounds of the problem.

        Args:
            bounds (Dict[str, Tuple[float, float]]): A dictionary of the form {parameter: (lower_bound, upper_bound)}
        """

        ProblemConstructor._checkBounds(bounds)

        for key, value in bounds.items():
            self._loweBounds.append(value[0])
            self._upperBounds.append(value[1])
            self.boundsExpressions.append(f"{value[0]} <= {key} <= {value[1]}")
            self.nvar += 1
            self.pnames.append(key)

    def setResults(self, expressions: Dict[str, Union[None, str]]) -> None:
        """Set the results to get from the simulation.

        Args:
            expressions (List[str]): List of results name as set in the FreeCAD spreadsheet.
        """
        ProblemConstructor._checkExpressions(expressions)  # type: ignore

        self.resultsExpressions = list(expressions.keys())
        self.iterableOutput = list(expressions.values())

    def getObjectives(self) -> List[Callable[(...), float]]:
        """Returns a list of callables to evaluate the objectives of the problem.

        Returns:
            List[Callable[(...), float]]: List of callables to evaluate the objectives of the problem.
        """
        return self._objectives

    def getConstraints(self) -> List[Callable[(...), float]]:
        """Returns a list of callables to evaluate the constraints of the problem.

        Returns:
            List[Callable[(...), float]]: List of callables to evaluate the constraints of the problem.
        """
        return self._constraints

    def getBounds(self) -> Tuple[List[float], List[float]]:
        """Returns the lower and upper bounds of the problem.

        Returns:
            Tuple[List[float], List[float]]: List of lower bounds and list of upper bounds.
        """
        return self._loweBounds, self._upperBounds

    def getResultsExpressions(self) -> List[str]:
        """Returns the names of the results from the simulation.

        Returns:
            List[str]: List of names of results.
        """
        return self.resultsExpressions

    def getIterableOutput(self) -> List[Union[None, str]]:
        """Specify how to treat iterable results from the simulation.

        Returns:
            List[str]: List of actions to perform on iterable results.
        """
        return self.iterableOutput
    
    def getObjectiveWeights(self) -> List[float]:
        """Returns the weights of the objectives.

        Returns:
            List[float]: List of weights of objectives.
        """
        return self.objectiveWeights
    
    def getConstraintsRelaxation(self) -> List[float]:
        """Returns the relaxation of the constraints.

        Returns:
            List[float]: List of relaxation of constraints.
        """
        return self.constraintsRelaxation

    def getPnames(self) -> List[str]:
        """Returns the names of the parameters.

        Returns:
            List[str]: The parameters names.
        """
        return self.pnames

    def getNobj(self) -> int:
        """Returns the number of objectives.

        Returns:
            int: The number of objectives.
        """
        return self.nobj

    def getNconst(self) -> int:
        """Returns the number of constraints.

        Returns:
            int: The number of constraints.
        """
        return self.nconst

    def getNvar(self) -> int:
        """Returns the number of variables.

        Returns:
            int: The number of variables.
        """
        return self.nvar

    def getObjectivesExpressions(self) -> List[str]:
        """Returns the list of objectives expressions.

        Returns:
            List[str]: A list of objectives expressions.
        """
        return self.objectivesExpressions

    def getConstraintsExpressions(self) -> List[str]:
        """Returns the list of constraints expressions.

        Returns:
            List[str]: A list of constraints expressions.
        """
        return self.constraintsExpressions

    @staticmethod
    def _expressionParser(expression: str) -> Tuple[List[str], List[str]]:
        """Parse an expression into a list of operands and a list of operations.

        Args:
            expression (str): The expression to parse.

        Returns:
            Tuple[List[str], List[str]]: A tuple of (operands, operations)

        Merit:
            This function is based on the following merit:
            Stack Overflow: https://stackoverflow.com/questions/13055884/parsing-math-expression-in-python-and-solving-to-find-an-answer
            Author: https://stackoverflow.com/users/748858/mgilson
        """
        expression = expression.replace(" ", "")
        operators = set("^+-*/")
        operations = (
            []
        )  # This holds the operators that are found in the string (left to right)
        operands = (
            []
        )  # this holds the non-operators that are found in the string (left to right)
        buff = []
        for c in expression:  # examine 1 character at a time
            if c in operators:
                # found an operator.  Everything we've accumulated in `buff` is
                # a single "number". Join it together and put it in `operands`.
                operands.append("".join(buff))
                buff = []
                operations.append(c)
            else:
                # not an operator.  Just accumulate this character in buff.
                buff.append(c)
        operands.append("".join(buff))
        return operands, operations

    @staticmethod
    def _expressionEvaluator(
        results: Dict[str, float], operands: List[str], operations: List[str]
    ) -> float:
        """Evaluate an expression given its operands and operations.

        Args:
            operands (List[str]): List of operands in the expression.
            operations (List[str]): List of operations in the expression.
            results (Dict[str, float]): Dictionary of results from the simulator.

        Raises:
            ValueError: When a operand used in the expression is not known.

        Returns:
            float: Result of the expression.

        Merit:
            This function is based on the following merit:
            Stack Overflow: https://stackoverflow.com/questions/13055884/parsing-math-expression-in-python-and-solving-to-find-an-answer
            Author: https://stackoverflow.com/users/748858/mgilson
        """

        operands_values = []
        for operand in operands:
            if operand in results:
                operands_values.append(results[operand])
            elif operand == "":
                operands_values.append(0)
            elif ProblemConstructor._testFloat(operand):
                operands_values.append(float(operand))
            else:
                raise ValueError(f"Unknown operand {operand}")

        operator_order = (
            "^",
            "*/",
            "+-",
        )  # precedence from left to right.  operators at same index have same precendece.
        # map operators to functions.
        op_dict = {
            "*": operator.mul,
            "/": operator.truediv,
            "+": operator.add,
            "-": operator.sub,
            "^": operator.pow,
        }

        operations_copy = operations.copy()

        for op in operator_order:  # Loop over precedence levels
            while any(
                o in operations_copy for o in op
            ):  # Operator with this precedence level exists
                idx, oo = next(
                    (i, o) for i, o in enumerate(operations_copy) if o in op
                )  # Next operator with this precedence
                operations_copy.pop(idx)  # remove this operator from the operator list
                values = map(
                    float, operands_values[idx : idx + 2]
                )  # here I just assume float for everything
                value = op_dict[oo](*values)
                operands_values[idx : idx + 2] = [value]  # clear out those indices

        return operands_values[0]

    @staticmethod
    def _checkExpressions(expressions: Dict[str, float]) -> bool:
        """_summary_

        Args:
            expression (Dict[str, float]): Expressions to be checked.

        Raises:
            TypeError: If any expressions are not encoded as strings.

        Returns:
            bool: True if all expressions are encoded as strings.
        """
        if all([isinstance(exp, str) for exp in expressions]):
            return True
        else:
            raise TypeError("The expressions must be encoded as strings.")

    @staticmethod
    def _checkBounds(bounds: Dict[str, Tuple[float, float]]) -> bool:  # type: ignore
        """_summary_

        Args:
            bounds (Dict[str, Tuple[float, float]]): Bounds to be checked.

        Raises:
            TypeError: If the bounds are not encoded as a dictionary.

        Returns:
            bool: True if the bounds are encoded as a dictionary.
        """
        if isinstance(bounds, dict):
            for name, value in bounds.items():
                name_check = isinstance(name, str)
                value_check = isinstance(value, Iterable)
                number_check = all([isinstance(num, (int, float)) for num in value])
                dim_check = value[0] < value[1]

                if not dim_check:
                    raise ValueError(
                        "The upper bound must be greater than the lower bound."
                    )
                elif not number_check:
                    raise TypeError("The bounds values must be integers or floats.")
                elif not value_check:
                    raise TypeError("The bounds values must be encoded as tuples.")
                elif not name_check:
                    raise TypeError("The bounds keys must be strings.")
                else:
                    return True
        else:
            raise TypeError("The bounds must be encoded as a dictionary.")

    @staticmethod
    def _testFloat(x: str) -> bool:
        """Test if a string can be converted to a float.

        Args:
            x (str): String to be tested.

        Returns:
            bool: True if the string can be converted to a float, False otherwise.
        """
        try:
            float(x)
            return True
        except ValueError:
            return False



from typing import Union

from pandas import DataFrame

from theeng.algorithms.visualizations import Visualizations
from theeng.core.abstract import Step


class Visualization(Step):
    def __init__(self, data: DataFrame):
        self.data = data

    def plot(
        self,
        visualizationName: str = "parallelCoordinate",
        savePath: Union[None, str] = None,
        **kwargs
    ):
        visualizationMethod = self._getMethod(
            Visualizations, visualizationName, data=self.data
        )(**kwargs)
        if savePath:
            visualizationMethod.write_html(savePath)



from typing import Callable, Dict, Iterable, List, Tuple

from numpy import concatenate
from pandas import DataFrame
from pymoo.core.callback import Callback
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize

from theeng.algorithms.optimizers import Optimizers
from theeng.core.abstract import Step
from theeng.core.problem import ProblemConstructor


class Optimizer(Step):
    def __init__(
        self,
        problem: ProblemConstructor,
        evaluator: Callable[[Dict[str, float]], Dict[str, float]],
    ) -> None:
        super().__init__(problem, evaluator)

    def optimize(
        self,
        optimizerName: str = "nsga3",
        termination: Tuple[str, int] = ("n_eval", 100),
        **kwargs
    ) -> Tuple[List[List[float]], List[List[float]], DataFrame]:
        if self.nObj > 1:
            if not optimizerName == "nsga3":
                raise Exception(
                    "Only NSGA3 is supported for multi-objective optimization. Use nsga3 name."
                )

        problem = OptimizationProblem(
            self.problem,
            self.evaluator,
        )
        algorithm = self._getMethod(Optimizers, optimizerName)(**kwargs, nObj=self.nObj)

        res = minimize(
            problem,
            algorithm,
            termination=termination,
            seed=1,
            callback=HistCallback(),
            return_least_infeasible=True,
        )

        x = res.X.tolist()
        f = res.F.tolist()

        if not isinstance(x[0], Iterable):
            x = [x]
        if not isinstance(f[0], Iterable):
            f = [f]

        x_hist = concatenate(res.algorithm.callback.data["x_hist"]).tolist()
        r_hist = concatenate(res.algorithm.callback.data["r_hist"]).tolist()

        data = concatenate([x_hist, r_hist], axis=1)
        data = DataFrame(
            data,
            columns=self.pNames
            + self.resultsExpressions
            + self.objectiveExpressions
            + self.constraintExpressions,
        )

        data = data.T.drop_duplicates().T  # drop duplicate columns

        return x, f, data

    def convertToSimulator(
        self,
        x: List[List[float]],
        simulator: Callable[[Dict[str, float]], Dict[str, float]],
    ) -> Tuple[List[List[float]], List[List[float]], DataFrame]:
        f = []
        r = []

        for design in x:
            parameters = {name: value for name, value in zip(self.pNames, design)}
            results = simulator(parameters)

            objs = [obj(results) for obj in self.objectives]
            consts = [constr(results) for constr in self.constraints]
            res = list(results.values()) + objs + consts

            f.append(objs)
            r.append(res)

        data = concatenate([x, r], axis=1)
        data = DataFrame(
            data,
            columns=self.pNames
            + self.resultsExpressions
            + self.objectiveExpressions
            + self.constraintExpressions,
        )

        data = data.T.drop_duplicates().T

        return x, f, data


class OptimizationProblem(ElementwiseProblem):
    def __init__(
        self,
        problem: ProblemConstructor,
        evaluator: Callable[[Dict[str, float]], Dict[str, float]],
        **kwargs
    ):
        """Initialize the optimization problem.

        Args:
            problem (ProblemConstructor): The problem to be evaluated.
            evaluator (Evaluator): The evaluator to be used.
        """

        self._evaluator = evaluator

        self._nvar = problem.getNvar()
        self._nobj = problem.getNobj()
        self._nconst = problem.getNconst()
        self._pnames = problem.getPnames()

        self._objectives = problem.getObjectives()
        self._constraints = problem.getConstraints()
        self._lowerBounds, self._upperBounds = problem.getBounds()

        super().__init__(
            n_var=self._nvar,
            n_obj=self._nobj,
            n_constr=self._nconst,
            xl=self._lowerBounds,
            xu=self._upperBounds,
        )

    def _evaluate(self, x, out: dict, *args, **kwargs):
        """Evaluate the optimization problem on the given designs.

        Args:
            x (_type_): Design samples.
            out (dict): dictionary containing the evaluated samples, objectives and constraints.
        """

        parameters = {name: value for name, value in zip(self._pnames, x)}
        results = self._evaluator(parameters)

        f = [obj(results) for obj in self._objectives]
        g = [constr(results) for constr in self._constraints]
        r = list(results.values())

        out["F"] = f
        out["G"] = g
        out["R"] = r + f + g


class HistCallback(Callback):
    """A class to store the all history of the optimization process."""

    def __init__(self) -> None:
        super().__init__()
        self.data["x_hist"] = []
        self.data["r_hist"] = []

    def notify(self, algorithm):
        self.data["x_hist"].append(algorithm.pop.get("X"))
        self.data["r_hist"].append(algorithm.pop.get("R"))



from typing import Union

from pymoo.algorithms.moo.unsga3 import NSGA3
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.algorithms.soo.nonconvex.nelder import NelderMead
from pymoo.algorithms.soo.nonconvex.pso import PSO
from pymoo.core.population import Population
from pymoo.factory import get_reference_directions
from pymoo.operators.sampling.rnd import FloatRandomSampling


class Optimizers:
    def __init__(self) -> None:
        pass

    def nelderMead(self, **kwargs):
        return NelderMead()

    def geneticAlgorithm(
        self,
        popSize: int,
        restartPop: Union[FloatRandomSampling, Population] = FloatRandomSampling(),
        **kwargs
    ):
        return GA(pop_size=popSize, eliminate_duplicates=True, sampling=restartPop)  # type: ignore

    def particleSwarm(
        self,
        popSize: int,
        restartPop: Union[FloatRandomSampling, Population] = FloatRandomSampling(),
        **kwargs
    ):
        return PSO(pop_size=popSize, sampling=restartPop)  # type: ignore

    def nsga3(
        self,
        popSize: int,
        nObj: int,
        restartPop: Union[FloatRandomSampling, Population] = FloatRandomSampling(),
        **kwargs
    ):
        ref_dirs = get_reference_directions("energy", nObj, n_points=nObj + 1, seed=1)
        return NSGA3(pop_size=popSize, ref_dirs=ref_dirs, eliminate_duplicates=True, sampling=restartPop)  # type: ignore



from scipy.stats.qmc import LatinHypercube


class Samplers:
    def __init__(self, nVar) -> None:
        self.nVar = nVar

    def latinHypercube(self) -> LatinHypercube:
        sampler = LatinHypercube(d=self.nVar)
        return sampler



from math import sqrt
from typing import Tuple

from pandas import DataFrame, concat

from theeng.core.problem import ProblemConstructor


class Rankers:
    def __init__(
        self, problem: ProblemConstructor, data: DataFrame, weights: Tuple[float]
    ) -> None:
        self.data = data
        self.objectivesData = data[problem.getObjectivesExpressions()]
        normData = Rankers._normalization(self.objectivesData)
        self.weightedNormData = Rankers._weightening(normData, weights)

    def topsis(self):
        bestDesign = self.weightedNormData.min(axis=0)
        worstDesign = self.weightedNormData.max(axis=0)
        performanceScore = []
        for _, design in self.weightedNormData.iterrows():
            positiveSeparation = sqrt(sum((design - bestDesign) ** 2))
            negativeSeparation = sqrt(sum((design - worstDesign) ** 2))
            performanceScore.append(
                negativeSeparation / (negativeSeparation + positiveSeparation)
            )
        resultData = concat(
            [
                self.data.reset_index(drop=True),
                DataFrame(
                    performanceScore,
                    columns=[
                        "Score",
                    ],
                ),
            ],
            axis=1,
        )
        sortedResultData = resultData.sort_values(
            "Score", ascending=True
        )  # sure is ascending?
        sortedResultData = Rankers._returnEfficient(sortedResultData, reverse=True) # topsis score is better when 1
        return sortedResultData

    def simpleAdditive(self):
        performanceScore = self.weightedNormData.sum(axis=1)
        resultData = concat(
            [
                self.data,
                DataFrame(
                    performanceScore,
                    columns=[
                        "Score",
                    ],
                ),
            ],
            axis=1,
        )
        sortedResultData = resultData.sort_values("Score", ascending=True)
        sortedResultData = Rankers._returnEfficient(sortedResultData, reverse=False) # simple additive score is better when 0
        return sortedResultData

    @staticmethod
    def _returnEfficient(data: DataFrame, efficiencyCliff: float = 0.20, reverse: bool = False):
        if "Score" in data.columns:
            sortedScores = sorted(data["Score"], reverse=reverse)
            numElementsToExtract = int(len(data["Score"]) * efficiencyCliff)  # get number of elements corresponding to the smaller 20%
            smallElements = sortedScores[:numElementsToExtract]  # get 20% smaller elements to extract

            data["Efficiency"] = [
                True if value in smallElements else False for value in data["Score"]
            ]
            return data
        return data

    @staticmethod
    def _normalization(data: DataFrame):
        normData = (data - data.min()) / (data.max() - data.min())
        return normData

    @staticmethod
    def _weightening(data: DataFrame, weights: Tuple[float]):
        weightedNormData = data.multiply(weights, axis=1)
        return weightedNormData



from setuptools import setup, find_packages

VERSION = "1.0"
DESCRIPTION = "General structure optimizer leveraging surrogates."


with open("requirements.txt") as f:
    required = f.read().splitlines()

# Setting up
setup(
    name="theeng",
    version=VERSION,
    author="Massimo Brivio",
    author_email="m.g.brivio@gmail.com",
    description=DESCRIPTION,
    packages=find_packages(where=".", include=["theeng*"]),
    long_description_content_type="text/markdown",
    long_description="TheEng is a Python library for dealing with structural optimization.",
    install_requires=required,
    url="https://github.com/massimobrivio/TheEng",
)

# 1 python setup.py sdist bdist
# 2 twine upload --repository testpypi dist/*
# 3 twine upload dist/* --> only for final release



from pandas import DataFrame
import plotly.express as px
from plotly.graph_objects import Figure


class Visualizations:
    def __init__(self, data: DataFrame) -> None:
        self.data = data

    def scatterPlot(self, **kwargs) -> Figure:
        xName = kwargs.get("xName")
        yName = kwargs.get("yName")
        if not (xName and yName):
            raise ValueError("You must specify xName and yName")

        if xName in self.data.columns and yName in self.data.columns:
            if "Efficiency" in self.data.columns:
                visualizationObject = px.scatter(
                    self.data, x=xName, y=yName, color="Efficiency"
                )
            else:
                visualizationObject = px.scatter(self.data, x=xName, y=yName)
        else:
            raise ValueError("xName or yName are not present in data.")
        return visualizationObject

    def heatMap(self, **kwargs) -> Figure:
        filteredData = self._filterData(**kwargs)
        visualizationObject = px.imshow(filteredData.corr(), text_auto=".1f")  # type: ignore
        return visualizationObject

    def parallelCoordinate(self, **kwargs) -> Figure:
        filteredData = self._filterData(**kwargs)
        visualizationObject = px.parallel_coordinates(
            filteredData, dimensions=list(filteredData.columns)
        )
        return visualizationObject

    def _filterData(self, **kwargs):
        columnsNames = kwargs.get("columnsNames")
        if columnsNames:
            filteredData = self.data[columnsNames]
        else:
            print("Using all data for visualization...")
            filteredData = self.data
        return filteredData



import numpy as np
from pandas import DataFrame

from theeng.algorithms.rankers import Rankers
from theeng.core.abstract import Step
from theeng.core.problem import ProblemConstructor


class Ranker(Step):
    def __init__(self, problem: ProblemConstructor, data: DataFrame) -> None:
        objectiveExpressions = problem.getObjectivesExpressions()
        constraintsExpressions = problem.getConstraintsExpressions()
        objectiveWeights = problem.getObjectiveWeights()
        constraintsRelaxation = problem.getConstraintsRelaxation()

        if not len(objectiveWeights) == len(objectiveExpressions):
            raise ValueError("Weights should be the same length as objectives.")

        if not sum(objectiveWeights) == 1:
            raise ValueError("Weights should sum up to 1.")

        if not len(constraintsRelaxation) == len(constraintsExpressions):
            raise ValueError(
                "Constraints relaxation should be the same length as constraints expressions."
            )

        elif constraintsRelaxation is None:
            constraintsRelaxation = [np.inf] * len(constraintsExpressions)

        for i in range(len(constraintsRelaxation)):
            minConstraintViolation = np.min(data[constraintsExpressions[i]])
            if constraintsRelaxation[i] < minConstraintViolation:
                print(
                    "Warning: Relaxation value is smaller than the minimum value of the constraint expression."
                )
                print(
                    f"Relaxation value will be set to the minimum value of the constraint expression -> {minConstraintViolation}"
                )
                constraintsRelaxation[i] = minConstraintViolation

        # get constraints expressions for which relaxation is not None
        constraintsRelaxationExpressions = [
            constraintsExpressions[i]
            for i in range(len(constraintsExpressions))
            if constraintsRelaxation[i] is not None
        ]
        # filter data with different conditions for each constraint column
        data = data[data[constraintsRelaxationExpressions].apply(lambda x: all(x <= constraintsRelaxation[i] for i, x in enumerate(x)), axis=1)]  # type: ignore

        self.problem = problem
        self.data = data
        self.objectiveWeights = objectiveWeights

    def rank(self, rankingName: str = "topsis"):
        rankingMethod = self._getMethod(
            Rankers,
            rankingName,
            problem=self.problem,
            data=self.data,
            weights=self.objectiveWeights,
        )
        data = rankingMethod()
        return data



from os.path import isfile
from pickle import dump, load
from typing import Callable, Dict, Tuple

from numpy import ndarray
from pandas import DataFrame
from sklearn.model_selection import cross_val_score

from theeng.algorithms.surrogates import Surrogates
from theeng.core.abstract import Step
from theeng.core.problem import ProblemConstructor


class Surrogate(Step):
    def __init__(
        self,
        problem: ProblemConstructor,
        data: DataFrame,
    ) -> None:
        parameterNames = problem.getPnames()
        resultsExpressions = problem.getResultsExpressions()

        self.trainingData_x = data[parameterNames].values
        self.trainingData_y = data[resultsExpressions].values
        self.trainedSurrogate = None
        self.resultsExpressions = resultsExpressions

    def generate(
        self, surrogateName: str = "polynomial", save: bool = False, **kwargs
    ) -> Tuple[Callable[[Dict[str, float]], Dict[str, float]], Tuple[float, float]]:
        surrogateMethod = self._getMethod(Surrogates, surrogateName)(**kwargs)
        trainedSurrogate, surrogatePerformance = self._train(
            surrogateMethod, save=save, **kwargs
        )
        self.trainedSurrogate = trainedSurrogate

        return self._predict, surrogatePerformance

    def generateFromFile(self, surrogatePath: str) -> Callable[(...), Dict[str, float]]:
        try:  # Try to load surrogate from file
            Surrogate._checkPath(
                surrogatePath,
                "No path to surrogate file specified. Please specify a path to load the surrogate.",
            )
            print(f"Trying to load surrogate from file... {surrogatePath}")
            trainedSurrogate = load(open(surrogatePath, "rb"))
            self.trainedSurrogate = trainedSurrogate
        except FileNotFoundError:
            raise ValueError(
                "No surrogate has been generated. Run method generate_surrogate first."
            )
        return self._predict

    def _predict(self, parameters: Dict[str, float]) -> Dict[str, float]:
        """Method to evaluate the surrogate model.

        Args:
            parameters (Dict[str, float]): A dicttionary of design parameters values and their aliases contained in the spreadsheet (names).

        Raises:
            ValueError: If no surrogate has been generated.

        Returns:
            Dict[str, float]: A dictionary containing results aliases and values.
        """
        if not self.trainedSurrogate:
            raise ValueError(
                "No surrogate has been generated. Use train() method first."
            )
        predictions = self.trainedSurrogate.predict([list(parameters.values())])  # type: ignore
        results = dict(zip(self.resultsExpressions, predictions[0]))
        return results

    def _train(
        self,
        surrogateMethod,
        save: bool = False,
        **kwargs,
    ) -> Tuple[object, Tuple[float, float]]:
        trainedSurrogate = surrogateMethod.fit(self.trainingData_x, self.trainingData_y)

        n_data_rows = len(self.trainingData_x)
        test_set_numdata = (
            n_data_rows * 0.2
        )  # 20% of the data is used for testing in cross validation.
        n_kfold_splits = (
            round(n_data_rows / test_set_numdata) if test_set_numdata > 2 else 2
        )
        scores = cross_val_score(
            trainedSurrogate,
            self.trainingData_x,
            self.trainingData_y,
            cv=n_kfold_splits,
        )
        surrogatePerformance = (scores.mean(), scores.std())

        if save:
            if not kwargs.get("surrogatePath"):
                raise ValueError(
                    "No path specified. Please specify a path to save the surrogate."
                )
            with open(kwargs.get("surrogatePath"), "wb") as f:  # type: ignore
                dump(trainedSurrogate, f)
            f.close()

        return trainedSurrogate, surrogatePerformance

    def getTrainingData(self) -> Tuple[ndarray, ndarray]:
        return self.trainingData_x, self.trainingData_y

    @staticmethod
    def _checkPath(path: str, *args) -> None:
        if not isfile(path):
            raise FileNotFoundError(args[0])
